from enum import StrEnum


class Messages(StrEnum):
    UNKNOWN_ERROR = "Something went wrong! ohh no... 😭"
    NO_ANSWER = "I Rather not answer that honestly..."
    FORBIDDEN_ACTION_CHANNEL = "I cannot do that in this channel..."
    NOT_IMPLEMENTED = "%s is not implemented for %s"
